import windows;

public:

typedef void* HDC;
typedef void* HGLRC;
typedef void* HGDIOBJ;
typedef void* HFONT;
typedef void* HPEN;
typedef void* HBRUSH;
typedef void* HBITMAP;
typedef void* HMENU;
typedef void* HICON;
typedef void* HCURSOR;
typedef void* ATOM;

alias uint COLORREF;
alias uint TCOLORREF;

struct RECT
{
  int left,top,right,bottom;
}
alias RECT* LPRECT;

struct POINT
{
  int px,py;
}
alias POINT* LPPOINT;

struct SIZE
{
  int cx,cy;
}
alias SIZE* LPSIZE;

// Pixel format descriptor
// I need to be more careful with struct alignment here!
struct PIXELFORMATDESCRIPTOR
{
  WORD  nSize;
  WORD  nVersion;
  DWORD dwFlags;
  BYTE  iPixelType;
  BYTE  cColorBits;
  BYTE  cRedBits;
  BYTE  cRedShift;
  BYTE  cGreenBits;
  BYTE  cGreenShift;
  BYTE  cBlueBits;
  BYTE  cBlueShift;
  BYTE  cAlphaBits;
  BYTE  cAlphaShift;
  BYTE  cAccumBits;
  BYTE  cAccumRedBits;
  BYTE  cAccumGreenBits;
  BYTE  cAccumBlueBits;
  BYTE  cAccumAlphaBits;
  BYTE  cDepthBits;
  BYTE  cStencilBits;
  BYTE  cAuxBuffers;
  BYTE  iLayerType;
  BYTE  bReserved;
  DWORD dwLayerMask;
  DWORD dwVisibleMask;
  DWORD dwDamageMask;
};
alias PIXELFORMATDESCRIPTOR* PPIXELFORMATDESCRIPTOR;
alias PIXELFORMATDESCRIPTOR* LPPIXELFORMATDESCRIPTOR;

// pixel types
const int PFD_TYPE_RGBA        = 0;
const int PFD_TYPE_COLORINDEX  = 1;

// layer types
const int PFD_MAIN_PLANE       = 0;
const int PFD_OVERLAY_PLANE    = 1;
const int PFD_UNDERLAY_PLANE   = -1;

// PIXELFORMATDESCRIPTOR flags
const int PFD_DOUBLEBUFFER            = 0x00000001;
const int PFD_STEREO                  = 0x00000002;
const int PFD_DRAW_TO_WINDOW          = 0x00000004;
const int PFD_DRAW_TO_BITMAP          = 0x00000008;
const int PFD_SUPPORT_GDI             = 0x00000010;
const int PFD_SUPPORT_OPENGL          = 0x00000020;
const int PFD_GENERIC_FORMAT          = 0x00000040;
const int PFD_NEED_PALETTE            = 0x00000080;
const int PFD_NEED_SYSTEM_PALETTE     = 0x00000100;
const int PFD_SWAP_EXCHANGE           = 0x00000200;
const int PFD_SWAP_COPY               = 0x00000400;
const int PFD_SWAP_LAYER_BUFFERS      = 0x00000800;
const int PFD_GENERIC_ACCELERATED     = 0x00001000;

// PIXELFORMATDESCRIPTOR flags for use in ChoosePixelFormat only
const int PFD_DEPTH_DONTCARE          = 0x20000000;
const int PFD_DOUBLEBUFFER_DONTCARE   = 0x40000000;
const int PFD_STEREO_DONTCARE         = 0x80000000;

enum
{
  DRIVERVERSION= 0,     /* Device driver version                    */
  TECHNOLOGY   = 2,     /* Device classification                    */
  HORZSIZE     = 4,     /* Horizontal size in millimeters           */
  VERTSIZE     = 6,     /* Vertical size in millimeters             */
  HORZRES      = 8,     /* Horizontal width in pixels               */
  VERTRES      = 10,    /* Vertical height in pixels                */
  BITSPIXEL    = 12,    /* Number of bits per pixel                 */
  PLANES       = 14,    /* Number of planes                         */
  NUMBRUSHES   = 16,    /* Number of brushes the device has         */
  NUMPENS      = 18,    /* Number of pens the device has            */
  NUMMARKERS   = 20,    /* Number of markers the device has         */
  NUMFONTS     = 22,    /* Number of fonts the device has           */
  NUMCOLORS    = 24,    /* Number of colors the device supports     */
  PDEVICESIZE  = 26,    /* Size required for device descriptor      */
  CURVECAPS    = 28,    /* Curve capabilities                       */
  LINECAPS     = 30,    /* Line capabilities                        */
  POLYGONALCAPS= 32,    /* Polygonal capabilities                   */
  TEXTCAPS     = 34,    /* Text capabilities                        */
  CLIPCAPS     = 36,    /* Clipping capabilities                    */
  RASTERCAPS   = 38,    /* Bitblt capabilities                      */
  ASPECTX      = 40,    /* Length of the X leg                      */
  ASPECTY      = 42,    /* Length of the Y leg                      */
  ASPECTXY     = 44,    /* Length of the hypotenuse                 */

  LOGPIXELSX  =  88,    /* Logical pixels/inch in X                 */
  LOGPIXELSY  =  90,    /* Logical pixels/inch in Y                 */

  SIZEPALETTE = 104,    /* Number of entries in physical palette    */
  NUMRESERVED = 106,    /* Number of reserved entries in palette    */
  COLORRES    = 108,    /* Actual color resolution                  */

  PHYSICALWIDTH  = 110, /* Physical Width in device units           */
  PHYSICALHEIGHT = 111, /* Physical Height in device units          */
  PHYSICALOFFSETX= 112, /* Physical Printable Area x margin         */
  PHYSICALOFFSETY= 113, /* Physical Printable Area y margin         */
  SCALINGFACTORX = 114, /* Scaling factor x                         */
  SCALINGFACTORY = 115, /* Scaling factor y                         */

  VREFRESH       = 116,  /* Current vertical refresh rate of the    */
                       /* display device (for displays only) in Hz*/
  DESKTOPVERTRES = 117,  /* Horizontal width of entire desktop in   */
                       /* pixels                                  */
  DESKTOPHORZRES = 118,  /* Vertical height of entire desktop in    */
                       /* pixels                                  */
  BLTALIGNMENT   = 119,  /* Preferred blt alignment                 */
}

extern(Windows)
{
  HDC GetDC(HWND);
  void ReleaseDC(HWND,HDC);
  int GetDeviceCaps(HDC, int);
  HGDIOBJ GetStockObject(int);
  int DescribePixelFormat(HDC, int, uint, LPPIXELFORMATDESCRIPTOR);
  int ChoosePixelFormat(HDC, LPPIXELFORMATDESCRIPTOR);
  int SetPixelFormat(HDC, uint, LPPIXELFORMATDESCRIPTOR);
  void SwapBuffers(HDC);
  HGLRC wglCreateContext(HDC);
  void wglDeleteContext(HGLRC);
  BOOL wglMakeCurrent(HDC, HGLRC);
}

